@@ -2,6 +2,7 @@ from __future__ import annotations
 
 import logging
 from enum import IntEnum, auto
+from inspect import signature
 from typing import Callable, Optional
 
 from .utils import _assert_same_length
@@ -9,7 +10,7 @@ from .utils import _assert_same_length
 logger = logging.getLogger(__name__)
 
 
-__version__ = "2.1.0"
+__version__ = "4.0.0"
 
 
 class DumpType(IntEnum):
@@ -21,12 +22,13 @@ class Node:
     def __init__(
         self,
         func: Callable,
-        name: Optional[str] = None,
         inputs: str | list[str] | None = None,
-        config: dict = None,
         outputs: str | list[str] | None = None,
+        name: Optional[str] = None,
+        config: dict = None,
         outputs_dumper: Callable | list[Callable] | None = None,
         outputs_dumper_type: DumpType = DumpType.INDIVIDUAL,
+        outputs_dumper_take_config: bool = False,
         outputs_path: Optional[list[str]] = None,
         outputs_loader: Callable | list[Callable] | None = None,
     ) -> None:
@@ -34,18 +36,19 @@ class Node:
         Parameters
         ------------------
         func: 適用する関数。
-        name: 関数の名前。Noneのときはfunc.__name__が代入される。__name__がなければ"anonymous"。
         inputs: 入力データ。パイプライン中の先頭ノードに対しNoneを設定すると、引数0個の関数をfuncにセットできる。
                 先頭ノード以外のものにNoneを設定した場合は、前段の出力を入力として使うという設定になる。
                 文字列が渡されているときは、dict型のPipiline.resultsからその名前の中間結果を読もうとする。
-        config: funcに与えるconfig、Noneでなければfuncの引数の最後にこれが加わる
         outputs: 出力結果をdict型のPipiline.resultsに入れる際のキー。Noneにすると保存されず、次のノードに渡されるのみとなる。
+        name: 関数の名前。Noneのときはfunc.__name__が代入される。__name__がなければ"anonymous"。
+        config: funcに与えるconfig、Noneでなければfuncの引数の最後にこれが加わる
         outputs_dumper: outputsをdumpする関数。リストを渡せば、各変数に対して別々の関数を適用可能。
                         引数は(data, filepath: str | PurePath) を想定。
                         outputs_dumper_typeがBATCHの際は、引数にデータのラベルを表すlist[str]が追加され、
                         (datalist, filepath: str | PurePath, datalabels)
                         となる。
         outputs_dumper_type: 複数の出力データを一つのファイルに保存したいときはDumpType.BATCHを設定する。
+        outputs_dumper_take_config: outputs_dumperがconfigを最後の引数にとるならTrueに設定する。
         outputs_path: dumpするファイルパス。
         outputs_loader: dumpしたoutputsをloadするための関数。
                         pipelineを途中から実行する場合、中間結果をloadする必要があるが、その際にコールされる。
@@ -65,6 +68,7 @@ class Node:
 
         self.outputs_dumper = outputs_dumper
         self.outputs_dumper_type = outputs_dumper_type
+        self.outputs_dumper_take_config = outputs_dumper_take_config
         if outputs_dumper_type == DumpType.BATCH:
             if not callable(outputs_dumper):
                 raise ValueError(
@@ -98,11 +102,17 @@ class Pipeline:
         name_duplicate_counter = {}
 
         for idx, node in enumerate(self.nodes):
-            if node.name in self.name_to_idx:
-                name_duplicate_counter[node.name] += 1
-                node.name += f"_{name_duplicate_counter[node.name]}"
-            else:
+            if node.name not in self.name_to_idx:
                 name_duplicate_counter[node.name] = 1
+            else:
+                name_duplicate_counter[node.name] += 1
+                node.name += f"__{name_duplicate_counter[node.name]}__"
+                if node.name not in self.name_to_idx:
+                    name_duplicate_counter[node.name] = 1
+                else:
+                    raise ValueError(
+                        f"name: {node.name} is duplicated. Consider change name"
+                    )
             self.name_to_idx[node.name] = idx
             if node.outputs is None:
                 continue
@@ -111,85 +121,126 @@ class Pipeline:
                     raise ValueError(f"node.output {output} is not unique")
                 self.outputs_to_indexes[output] = (idx, idx_outputs)
 
+        for node in self.nodes[1:]:
+            if node.inputs is None:
+                continue
+            for input in node.inputs:
+                if input not in self.outputs_to_indexes:
+                    raise ValueError(
+                        f"inputs: {input} will not be calculated in this pipeline"
+                    )
+
     def get_node_names(self) -> list[str]:
         return [node.name for node in self.nodes]
 
-    def run(self, start: int | str = 0):
+    def run(self, from_: int | str = 0, to_: int | str | None = None):
         """pipelineを実行する。戻り値はlist。
         Parameters
         ----------------
         start: どのノードからパイプラインを開始するか。インデックスかnameで指定可能。
+        end: どのノードのまでパイプラインを実行するか。インデックスかnameで指定可能。
         """
-        idx_start = self._get_start_index(start)
+        if to_ is None:
+            to_ = len(self.nodes) - 1
+        idx_from = self._get_start_or_end_index(from_, "start")
+        self.idx_from = idx_from
+        idx_to = self._get_start_or_end_index(to_, "end")
+        if idx_from > idx_to:
+            raise ValueError(
+                f"idx_from must satisfy idx_from ({idx_from}) <= idx_to ({idx_to})"
+            )
+
         logger.info(
-            f"Total {len(self.nodes)} tasks, scheduled {len(self.nodes[idx_start:])} tasks"
+            f"Scheduled {len(self.nodes)} tasks, {len(self.nodes[idx_from:idx_to+1])} tasks in total"
         )
-        outputs = self._load_interim_output(idx_start)
-        for idx, node in enumerate(self.nodes[idx_start:]):
+        outputs = self._load_interim_output()
+        for idx, node in enumerate(self.nodes[idx_from : idx_to + 1]):
             logger.info(
-                f"Running {idx+1}/{len(self.nodes[idx_start:])} tasks ({node.name})"
+                f"Running {idx+1}/{len(self.nodes[idx_from:idx_to+1])} tasks ({node.name})"
             )
 
             # 最初のノードだけ特別扱い
             args = []
-            if idx == 0 and idx_start == 0:
+            if idx == 0 and idx_from == 0:
                 if node.inputs is not None:
                     args.extend(node.inputs)
             else:
-                args.extend(self._get_inputs_from_previous_node(node, outputs))
+                inputs = self._get_inputs_from_previous_node(node, outputs)
+                if inputs is not None:
+                    args.extend(inputs)
 
             if node.config is not None:
                 args.append(node.config)
             outputs = node.func(*args)
 
             outputs = _convert_item_to_list(outputs)
+            if outputs is None:
+                continue
             _assert_non_zero_length(outputs, "outputs")
             self._insert_outputs_to_dict(node, outputs)
             self._dump_outputs(node, outputs)
 
-        logger.info("All tasks completed")
+        logger.info("Completed all tasks")
         return outputs
 
-    def _get_start_index(self, start: int | str):
-        if isinstance(start, int):
-            idx_start = start
-        elif isinstance(start, str):
-            idx_start = self.name_to_idx[start]
+    def _get_start_or_end_index(self, start_or_end: int | str, start_or_end_str: str):
+        if isinstance(start_or_end, int):
+            idx = start_or_end
+        elif isinstance(start_or_end, str):
+            try:
+                idx = self.name_to_idx[start_or_end]
+            except KeyError:
+                raise ValueError(
+                    f"specified {start_or_end_str} node ({start_or_end}) is not found"
+                )
 
-        if idx_start < 0 or idx_start >= len(self.nodes):
+        if idx < 0 or idx >= len(self.nodes):
             raise ValueError(
-                f"Wrong argument: idx_start ({idx_start}) must satisfy 0 <= start < {len(self.node)-1}"
+                f"0 <= {start_or_end_str} ({idx}) <= {len(self.nodes)-1} must be satisfied"
             )
+        return idx
 
-        return idx_start
-
-    def _load_interim_output(self, idx_start):
+    def _load_interim_output(self):
         """
-        idx_startから最後のノードまでのノードに対し、
-        idx_startより前のノードの結果を使うかチェックし、
+        idx_fromから最後のノードまでのノードに対し、
+        idx_fromより前のノードの結果を使うかチェックし、
         もし使うのであれば必要な結果をロードする。
         """
-        if idx_start == 0:
+        if self.idx_from == 0:
             return None
         last_output = None
         loaded_files_set = set()
-        # idx_startから読み込むこと
+        # idx_fromから読み込むこと
         # さもないと直前のノードの入力が2回読まれる可能性がある
-        for idx_ in range(idx_start, len(self.nodes)):
+        for idx_ in range(self.idx_from, len(self.nodes)):
+            # 関数の引数がゼロ個の時は何もしない
+            num_param_when_no_inputs = 0 if self.nodes[idx_].config is None else 1
+            if (
+                len(signature(self.nodes[idx_].func).parameters)
+                == num_param_when_no_inputs
+            ):
+                continue
             # 入力が指定されていない場合（前段の出力を入力とする場合）
             if self.nodes[idx_].inputs is None:
-                # idx_start以降のタスクはまだ計算していないのでcontinue
-                if idx_ > idx_start:
+                # idx_from以降のタスクはまだ計算していないのでcontinue
+                if idx_ > self.idx_from:
                     continue
                 last_output = self._load_last_output(
-                    self.nodes[idx_ - 1], loaded_files_set
+                    self.nodes[idx_ - 1],
+                    loaded_files_set,
+                    idx_ - 1,
                 )
             else:  # 前段より前の出力を入力にする場合
-                self._load_past_output(self.nodes[idx_], idx_start, loaded_files_set)
+                self._load_past_output(self.nodes[idx_], loaded_files_set)
         return last_output
 
-    def _load_last_output(self, node_prev, loaded_files_set):
-        # 本関数はrange(idx_start, len(self.nodes)): のループの最初に実行されるので、
+    def _load_last_output(self, node_prev, loaded_files_set, idx):
+        if node_prev.outputs_loader is None:
+            raise ValueError(
+                f"outputs loader of node {node_prev.name} (index: {idx}) is not set."
+            )
+
+        # 本関数はrange(idx_from, len(self.nodes)): のループの最初に実行されるので、
         # outputs_loaderの呼び出しが冗長ではない
         if node_prev.outputs_loader_type == DumpType.BATCH:
             loaded_files_set.add(*node_prev.outputs_path)
@@ -213,10 +264,10 @@ class Pipeline:
             outputs.append(outputs_loader(output_path))
         return outputs
 
-    def _load_past_output(self, node, idx_start, loaded_files_set):
+    def _load_past_output(self, node, loaded_files_set):
         for input in node.inputs:
             idx_dependant_node, idx_in_outputs = self.outputs_to_indexes[input]
-            if idx_dependant_node >= idx_start:
+            if idx_dependant_node >= self.idx_from:
                 # まだ計算していないのでcontinue
                 continue
 
@@ -258,8 +309,8 @@ class Pipeline:
         """前段までの入力を基に、次のNodeへの入力を決める"""
         # 入力がない場合は前段の結果を持ってくる
         if node.inputs is None:
-            if previous_outputs is None:
-                raise ValueError("inputs is None")
+            # if previous_outputs is None:
+            #     raise ValueError("inputs is None")
             return previous_outputs
 
         _assert_non_zero_length(node.inputs, "node.inputs")
@@ -279,7 +330,10 @@ class Pipeline:
             return
 
         if node.outputs_dumper_type == DumpType.BATCH:
-            return node.outputs_dumper(outputs, *node.outputs_path, node.outputs)
+            args = [outputs, *node.outputs_path, node.outputs]
+            if node.outputs_dumper_take_config:
+                args.append(node.config)
+            return node.outputs_dumper(*args)
 
         _assert_same_length(outputs, node.outputs_path, "outputs", "node.outputs_path")
         outputs_dumpers = _convert_item_to_list(node.outputs_dumper, len(outputs))
@@ -288,7 +342,82 @@ class Pipeline:
             outputs, node.outputs_path, outputs_dumpers
         ):
             if outputs_dumper is not None:
-                outputs_dumper(output, filepath)
+                args = [output, filepath]
+                if node.outputs_dumper_take_config:
+                    args.append(node.config)
+                outputs_dumper(*args)
+
+
+class TrivialPipeline:
+    """
+    関数のリストを受け取って順次実行するだけのパイプライン
+    データの受渡しは行わない。
+    """
+
+    def __init__(self, funcs: list[Callable], names: list[str] = None) -> None:
+        """
+        Parameters
+        --------------
+        funcs: 呼び出される関数群
+        names: 関数の名前
+        """
+        self.funcs = funcs
+        if names is not None:
+            _assert_same_length(funcs, names, "funcs", "names")
+            if len(names) != len(set(names)):
+                raise ValueError(f"names is not unique: {names}")
+            self.names = names
+            return
+
+        self.names = []
+        name_duplicate_counter = {}
+
+        for func in self.funcs:
+            if hasattr(func, "__name__"):
+                name = func.__name__
+            else:
+                name = "anonymous"
+
+            dup_counter = name_duplicate_counter.get(name, 0)
+            if dup_counter == 0:
+                name_duplicate_counter[name] = 1
+            else:
+                if name in name_duplicate_counter:
+                    name_duplicate_counter[name] = dup_counter + 1
+                    name += f"__{dup_counter + 1}__"
+                    dup_counter = name_duplicate_counter.get(name, 0)
+                    if dup_counter == 0:
+                        name_duplicate_counter[name] = 1
+                    else:
+                        raise ValueError(
+                            f"name: {name} is duplicated. Consider change name"
+                        )
+            self.names.append(name)
+
+    def run(self, from_=0, to_=None):
+        if to_ is None:
+            to_ = len(self.funcs) - 1
+
+        if from_ < 0 or from_ >= len(self.funcs):
+            raise ValueError(
+                f"0 <= from({from_}) <= {len(self.funcs)-1} must be satisfied"
+            )
+        if from_ > to_:
+            raise ValueError("start <= to must be satisfied")
+        if to_ < 0 or to_ >= len(self.funcs):
+            raise ValueError(f"0 <= to({to_}) <= {len(self.funcs)-1} must be satisfied")
+
+        logger.info(
+            f"Scheduled {len(self.funcs[from_:to_+1])} tasks, {len(self.funcs)} tasks in total"
+        )
+        for idx, (func, name) in enumerate(
+            zip(self.funcs[from_ : to_ + 1], self.names[from_ : to_ + 1])
+        ):
+            logger.info(
+                f"Running {idx+1}/{len(self.funcs[from_:to_+1])} tasks ({name})"
+            )
+            func()
+        logger.info("Completed all tasks!")
 
 
 def _assert_non_zero_length(x, x_str):
